#!/usr/bin/env python

"""Performs pre-MEDS data wrangling for MIMIC-IV."""

from datetime import datetime
from functools import partial
from pathlib import Path

import hydra
import polars as pl
from loguru import logger
from omegaconf import DictConfig

from MEDS_transforms.extract.utils import get_supported_fp
from MEDS_transforms.utils import get_shard_prefix, hydra_loguru_init, write_lazyframe


def add_discharge_time_by_hadm_id(
    df: pl.LazyFrame, discharge_time_df: pl.LazyFrame, out_column_name: str = "hadm_discharge_time"
) -> pl.LazyFrame:
    """Joins the two dataframes by ``"hadm_id"`` and adds the discharge time to the original dataframe."""

    discharge_time_df = discharge_time_df.select("hadm_id", pl.col("dischtime").alias(out_column_name))
    return df.join(discharge_time_df, on="hadm_id", how="left")


def fix_static_data(raw_static_df: pl.LazyFrame, death_times_df: pl.LazyFrame) -> pl.LazyFrame:
    """Fixes the static data by adding the death time to the static data and fixes the DOB nonsense.

    Args:
        raw_static_df: The raw static data.
        death_times_df: The death times data.

    Returns:
        The fixed static data.
    """

    death_times_df = death_times_df.group_by("subject_id").agg(pl.col("deathtime").min())

    return raw_static_df.join(death_times_df, on="subject_id", how="left").select(
        "subject_id",
        pl.coalesce(pl.col("deathtime"), pl.col("dod")).alias("dod"),
        (pl.col("anchor_year") - pl.col("anchor_age")).cast(str).alias("year_of_birth"),
        "gender",
    )


FUNCTIONS = {
    "hosp/diagnoses_icd": (add_discharge_time_by_hadm_id, ("hosp/admissions", ["hadm_id", "dischtime"])),
    "hosp/drgcodes": (add_discharge_time_by_hadm_id, ("hosp/admissions", ["hadm_id", "dischtime"])),
    "hosp/patients": (fix_static_data, ("hosp/admissions", ["subject_id", "deathtime"])),
}


@hydra.main(version_base=None, config_path="configs", config_name="pre_MEDS")
def main(cfg: DictConfig):
    """Performs pre-MEDS data wrangling for MIMIC-IV.

    Inputs are the raw MIMIC files, read from the `raw_cohort_dir` config parameter. Output files are either
    symlinked (if they are not modified) or written in processed form to the `MEDS_input_dir` config
    parameter. Hydra is used to manage configuration parameters and logging.
    """

    hydra_loguru_init()

    raw_cohort_dir = Path(cfg.raw_cohort_dir)
    MEDS_input_dir = Path(cfg.output_dir)

    all_fps = list(raw_cohort_dir.glob("**/*.*"))

    dfs_to_load = {}
    seen_fps = {}

    for in_fp in all_fps:
        pfx = get_shard_prefix(raw_cohort_dir, in_fp)

        try:
            fp, read_fn = get_supported_fp(raw_cohort_dir, pfx)
        except FileNotFoundError:
            logger.info(f"Skipping {pfx} @ {str(in_fp.resolve())} as no compatible dataframe file was found.")
            continue

        if fp.suffix in [".csv", ".csv.gz"]:
            read_fn = partial(read_fn, infer_schema_length=100000)

        if str(fp.resolve()) in seen_fps:
            continue
        else:
            seen_fps[str(fp.resolve())] = read_fn

        out_fp = MEDS_input_dir / fp.relative_to(raw_cohort_dir)

        if out_fp.is_file():
            print(f"Done with {pfx}. Continuing")
            continue

        out_fp.parent.mkdir(parents=True, exist_ok=True)

        if pfx not in FUNCTIONS:
            logger.info(
                f"No function needed for {pfx}: " f"Symlinking {str(fp.resolve())} to {str(out_fp.resolve())}"
            )
            relative_in_fp = fp.relative_to(out_fp.resolve().parent, walk_up=True)
            out_fp.symlink_to(relative_in_fp)
            continue
        else:
            out_fp = MEDS_input_dir / f"{pfx}.parquet"
            if out_fp.is_file():
                print(f"Done with {pfx}. Continuing")
                continue

            fn, need_df = FUNCTIONS[pfx]
            if not need_df:
                st = datetime.now()
                logger.info(f"Processing {pfx}...")
                df = read_fn(fp)
                logger.info(f"  Loaded raw {fp} in {datetime.now() - st}")
                processed_df = fn(df)
                write_lazyframe(processed_df, out_fp)
                logger.info(f"  Processed and wrote to {str(out_fp.resolve())} in {datetime.now() - st}")
            else:
                needed_pfx, needed_cols = need_df
                if needed_pfx not in dfs_to_load:
                    dfs_to_load[needed_pfx] = {"fps": set(), "cols": set()}

                dfs_to_load[needed_pfx]["fps"].add(fp)
                dfs_to_load[needed_pfx]["cols"].update(needed_cols)

    for df_to_load_pfx, fps_and_cols in dfs_to_load.items():
        fps = fps_and_cols["fps"]
        cols = list(fps_and_cols["cols"])

        df_to_load_fp, df_to_load_read_fn = get_supported_fp(raw_cohort_dir, df_to_load_pfx)

        st = datetime.now()

        logger.info(f"Loading {str(df_to_load_fp.resolve())} for manipulating other dataframes...")
        if df_to_load_fp.suffix in [".csv.gz"]:
            df = read_fn(df_to_load_fp, columns=cols)
        else:
            df = read_fn(df_to_load_fp)
        logger.info(f"  Loaded in {datetime.now() - st}")

        for fp in fps:
            pfx = get_shard_prefix(raw_cohort_dir, fp)
            out_fp = MEDS_input_dir / f"{pfx}.parquet"

            logger.info(f"  Processing dependent df @ {pfx}...")
            fn, _ = FUNCTIONS[pfx]

            fp_st = datetime.now()
            logger.info(f"    Loading {str(fp.resolve())}...")
            fp_df = seen_fps[str(fp.resolve())](fp)
            logger.info(f"    Loaded in {datetime.now() - fp_st}")
            processed_df = fn(fp_df, df)
            write_lazyframe(processed_df, out_fp)
            logger.info(f"    Processed and wrote to {str(out_fp.resolve())} in {datetime.now() - fp_st}")

    logger.info(f"Done! All dataframes processed and written to {str(MEDS_input_dir.resolve())}")


if __name__ == "__main__":
    main()



"""Tests the tokenization script.

Note that this test relies on the normalized shards from the normalization test.

Set the bash env variable `DO_USE_LOCAL_SCRIPTS=1` to use the local py files, rather than the installed
scripts.
"""

from datetime import datetime

import polars as pl

from .test_normalization import NORMALIZED_MEDS_SCHEMA
from .test_normalization import WANT_SHARDS as NORMALIZED_SHARDS
from .transform_tester_base import TOKENIZATION_SCRIPT, single_stage_transform_tester

SECONDS_PER_DAY = 60 * 60 * 24


def ts_to_time_delta_days(ts: list[list[datetime]]) -> list[list[float]]:
    """TODO: Doctests"""
    out = []
    for patient_ts in ts:
        out.append([float("nan")])
        for i in range(1, len(patient_ts)):
            out[-1].append((patient_ts[i] - patient_ts[i - 1]).total_seconds() / SECONDS_PER_DAY)
    return out


# TODO: Make these schemas exportable, maybe???
# TODO: Why is the code getting converted to a float?
SCHEMAS_SCHEMA = {
    "patient_id": NORMALIZED_MEDS_SCHEMA["patient_id"],
    "code": pl.List(NORMALIZED_MEDS_SCHEMA["code"]),
    "numeric_value": pl.List(NORMALIZED_MEDS_SCHEMA["numeric_value"]),
    "start_time": NORMALIZED_MEDS_SCHEMA["time"],
    "time": pl.List(NORMALIZED_MEDS_SCHEMA["time"]),
}

SEQ_SCHEMA = {
    "patient_id": NORMALIZED_MEDS_SCHEMA["patient_id"],
    "code": pl.List(pl.List(pl.UInt8)),
    "numeric_value": pl.List(pl.List(NORMALIZED_MEDS_SCHEMA["numeric_value"])),
    "time_delta_days": pl.List(pl.Float32),
}

TRAIN_0_TIMES = [
    [
        datetime(1980, 12, 28),
        datetime(2010, 5, 11, 17, 41, 51),
        datetime(2010, 5, 11, 17, 48, 48),
        datetime(2010, 5, 11, 18, 25, 35),
        datetime(2010, 5, 11, 18, 57, 18),
        datetime(2010, 5, 11, 19, 27, 19),
    ],
    [
        datetime(1978, 6, 20),
        datetime(2010, 6, 20, 19, 23, 52),
        datetime(2010, 6, 20, 19, 25, 32),
        datetime(2010, 6, 20, 19, 45, 19),
        datetime(2010, 6, 20, 20, 12, 31),
        datetime(2010, 6, 20, 20, 24, 44),
        datetime(2010, 6, 20, 20, 41, 33),
        datetime(2010, 6, 20, 20, 50, 4),
    ],
]
WANT_SCHEMAS_TRAIN_0 = pl.DataFrame(
    {
        "patient_id": [239684, 1195293],
        "code": [[7, 9], [6, 9]],
        "numeric_value": [[None, 1.5770268440246582], [None, 0.06802856922149658]],
        "start_time": [ts[0] for ts in TRAIN_0_TIMES],
        "time": TRAIN_0_TIMES,
    },
    schema=SCHEMAS_SCHEMA,
)

WANT_EVENT_SEQ_TRAIN_0 = pl.DataFrame(
    {
        "patient_id": [239684, 1195293],
        "time_delta_days": ts_to_time_delta_days(TRAIN_0_TIMES),
        "code": [
            [[5], [1, 10, 11], [10, 11], [10, 11], [10, 11], [4]],
            [[5], [1, 10, 11], [10, 11], [10, 11], [10, 11], [10, 11], [10, 11], [4]],
        ],
        "numeric_value": [
            [
                [float("nan")],
                [float("nan"), -0.569736897945404, -1.2714673280715942],
                [-0.43754738569259644, -1.168027639389038],
                [0.001321975840255618, -1.37490713596344],
                [-0.04097883030772209, -1.5300706624984741],
                [float("nan")],
            ],
            [
                [float("nan")],
                [float("nan"), -0.23133166134357452, 0.7973543255932579],
                [0.03833488002419472, 0.7973543255932579],
                [0.3397272229194641, 0.745638906955719],
                [-0.046266332268714905, 0.6939135937995033],
                [-0.3000703752040863, 0.7973543255932579],
                [-0.31064537167549133, 1.004242181777954],
                [float("nan")],
            ],
        ],
    },
    schema=SEQ_SCHEMA,
)

TRAIN_1_TIMES = [
    [datetime(1978, 3, 9), datetime(2010, 5, 26, 2, 30, 56), datetime(2010, 5, 26, 4, 51, 52)],
    [datetime(1976, 3, 28), datetime(2010, 2, 5, 5, 55, 39), datetime(2010, 2, 5, 7, 2, 30)],
]

WANT_SCHEMAS_TRAIN_1 = pl.DataFrame(
    {
        "patient_id": [68729, 814703],
        "code": [[8, 9], [8, 9]],
        "numeric_value": [[None, -0.5438239574432373], [None, -1.1012336015701294]],
        "start_time": [ts[0] for ts in TRAIN_1_TIMES],
        "time": TRAIN_1_TIMES,
    },
    schema=SCHEMAS_SCHEMA,
)

WANT_EVENT_SEQ_TRAIN_1 = pl.DataFrame(
    {
        "patient_id": [68729, 814703],
        "time_delta_days": ts_to_time_delta_days(TRAIN_1_TIMES),
        "code": [[[5], [3, 10, 11], [4]], [[5], [2, 10, 11], [4]]],
        "numeric_value": [
            [[float("nan")], [float("nan"), -1.4474751949310303, -0.3404940366744995], [float("nan")]],
            [[float("nan")], [float("nan"), 3.0046675205230713, 0.8490786552429199], [float("nan")]],
        ],
    },
    schema=SEQ_SCHEMA,
)

TUNING_0_TIMES = [[datetime(1988, 12, 19), datetime(2010, 1, 3, 6, 27, 59), datetime(2010, 1, 3, 8, 22, 13)]]

WANT_SCHEMAS_TUNING_0 = pl.DataFrame(
    {
        "patient_id": [754281],
        "code": [[7, 9]],
        "numeric_value": [[None, 0.28697699308395386]],
        "start_time": [ts[0] for ts in TUNING_0_TIMES],
        "time": TUNING_0_TIMES,
    },
    schema=SCHEMAS_SCHEMA,
)

WANT_EVENT_SEQ_TUNING_0 = pl.DataFrame(
    {
        "patient_id": [754281],
        "time_delta_days": ts_to_time_delta_days(TUNING_0_TIMES),
        "code": [[[5], [3, 10, 11], [4]]],
        "numeric_value": [
            [[float("nan")], [float("nan"), 1.513569951057434, 0.6939190626144409], [float("nan")]],
        ],
    },
    schema=SEQ_SCHEMA,
)


HELD_OUT_0_TIMES = [
    [
        datetime(1986, 7, 20),
        datetime(2010, 6, 3, 14, 54, 38),
        datetime(2010, 6, 3, 15, 39, 49),
        datetime(2010, 6, 3, 16, 20, 49),
        datetime(2010, 6, 3, 16, 44, 26),
    ]
]

WANT_SCHEMAS_HELD_OUT_0 = pl.DataFrame(
    {
        "patient_id": [1500733],
        "code": [[7, 9]],
        "numeric_value": [[None, -0.7995940446853638]],
        "start_time": [ts[0] for ts in HELD_OUT_0_TIMES],
        "time": HELD_OUT_0_TIMES,
    },
    schema=SCHEMAS_SCHEMA,
)

WANT_EVENT_SEQ_HELD_OUT_0 = pl.DataFrame(
    {
        "patient_id": [1500733],
        "time_delta_days": ts_to_time_delta_days(HELD_OUT_0_TIMES),
        "code": [[[5], [2, 10, 11], [10, 11], [10, 11], [4]]],
        "numeric_value": [
            [
                [float("nan")],
                [float("nan"), -1.1619458198547363, 0.7973587512969971],
                [-1.5320764780044556, 0.9525222778320312],
                [-1.230684518814087, 0.8490786552429199],
                [float("nan")],
            ]
        ],
    },
    schema=SEQ_SCHEMA,
)


WANT_SCHEMAS = {
    "schemas/train/0": WANT_SCHEMAS_TRAIN_0,
    "schemas/train/1": WANT_SCHEMAS_TRAIN_1,
    "schemas/tuning/0": WANT_SCHEMAS_TUNING_0,
    "schemas/held_out/0": WANT_SCHEMAS_HELD_OUT_0,
}

WANT_EVENT_SEQS = {
    "event_seqs/train/0": WANT_EVENT_SEQ_TRAIN_0,
    "event_seqs/train/1": WANT_EVENT_SEQ_TRAIN_1,
    "event_seqs/tuning/0": WANT_EVENT_SEQ_TUNING_0,
    "event_seqs/held_out/0": WANT_EVENT_SEQ_HELD_OUT_0,
}


def test_tokenization():
    single_stage_transform_tester(
        transform_script=TOKENIZATION_SCRIPT,
        stage_name="tokenization",
        transform_stage_kwargs=None,
        input_shards=NORMALIZED_SHARDS,
        want_data={**WANT_SCHEMAS, **WANT_EVENT_SEQS},
    )



from importlib.metadata import PackageNotFoundError, version
from importlib.resources import files

import polars as pl

__package_name__ = "MEDS_transforms"
try:
    __version__ = version(__package_name__)
except PackageNotFoundError:
    __version__ = "unknown"

PREPROCESS_CONFIG_YAML = files(__package_name__).joinpath("configs/preprocess.yaml")
EXTRACT_CONFIG_YAML = files(__package_name__).joinpath("configs/extract.yaml")

MANDATORY_COLUMNS = ["patient_id", "time", "code", "numeric_value"]

MANDATORY_TYPES = {
    "patient_id": pl.Int64,
    "time": pl.Datetime("us"),
    "code": pl.String,
    "numeric_value": pl.Float32,
    "categorical_value": pl.String,
    "text_value": pl.String,
}

DEPRECATED_NAMES = {
    "numerical_value": "numeric_value",
    "categoric_value": "categoric_value",
    "category_value": "categoric_value",
    "textual_value": "text_value",
    "timestamp": "time",
    "subject_id": "patient_id",
}

INFERRED_STAGE_KEYS = {
    "is_metadata",
    "data_input_dir",
    "metadata_input_dir",
    "output_dir",
    "reducer_output_dir",
}



"""Base helper code and data inputs for all transforms integration tests.

Set the bash env variable `DO_USE_LOCAL_SCRIPTS=1` to use the local py files, rather than the installed
scripts.
"""

from yaml import load as load_yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

import json
import os
import tempfile
from collections import defaultdict
from contextlib import contextmanager
from io import StringIO
from pathlib import Path

import numpy as np
import polars as pl
import rootutils
from nested_ragged_tensors.ragged_numpy import JointNestedRaggedTensorDict

from .utils import MEDS_PL_SCHEMA, assert_df_equal, parse_meds_csvs, run_command

root = rootutils.setup_root(__file__, dotenv=True, pythonpath=True, cwd=True)

code_root = root / "src" / "MEDS_transforms"
transforms_root = code_root / "transforms"
filters_root = code_root / "filters"

if os.environ.get("DO_USE_LOCAL_SCRIPTS", "0") == "1":
    # Root Source
    AGGREGATE_CODE_METADATA_SCRIPT = code_root / "aggregate_code_metadata.py"
    FIT_VOCABULARY_INDICES_SCRIPT = code_root / "fit_vocabulary_indices.py"
    RESHARD_TO_SPLIT_SCRIPT = code_root / "reshard_to_split.py"

    # Filters
    FILTER_MEASUREMENTS_SCRIPT = filters_root / "filter_measurements.py"
    FILTER_PATIENTS_SCRIPT = filters_root / "filter_patients.py"

    # Transforms
    ADD_TIME_DERIVED_MEASUREMENTS_SCRIPT = transforms_root / "add_time_derived_measurements.py"
    REORDER_MEASUREMENTS_SCRIPT = transforms_root / "reorder_measurements.py"
    EXTRACT_VALUES_SCRIPT = transforms_root / "extract_values.py"
    NORMALIZATION_SCRIPT = transforms_root / "normalization.py"
    OCCLUDE_OUTLIERS_SCRIPT = transforms_root / "occlude_outliers.py"
    TENSORIZATION_SCRIPT = transforms_root / "tensorization.py"
    TOKENIZATION_SCRIPT = transforms_root / "tokenization.py"
else:
    # Root Source
    AGGREGATE_CODE_METADATA_SCRIPT = "MEDS_transform-aggregate_code_metadata"
    FIT_VOCABULARY_INDICES_SCRIPT = "MEDS_transform-fit_vocabulary_indices"
    RESHARD_TO_SPLIT_SCRIPT = "MEDS_transform-reshard_to_split"

    # Filters
    FILTER_MEASUREMENTS_SCRIPT = "MEDS_transform-filter_measurements"
    FILTER_PATIENTS_SCRIPT = "MEDS_transform-filter_patients"

    # Transforms
    ADD_TIME_DERIVED_MEASUREMENTS_SCRIPT = "MEDS_transform-add_time_derived_measurements"
    REORDER_MEASUREMENTS_SCRIPT = "MEDS_transform-reorder_measurements"
    EXTRACT_VALUES_SCRIPT = "MEDS_transform-extract_values"
    NORMALIZATION_SCRIPT = "MEDS_transform-normalization"
    OCCLUDE_OUTLIERS_SCRIPT = "MEDS_transform-occlude_outliers"
    TENSORIZATION_SCRIPT = "MEDS_transform-tensorization"
    TOKENIZATION_SCRIPT = "MEDS_transform-tokenization"

# Test MEDS data (inputs)

SHARDS = {
    "train/0": [239684, 1195293],
    "train/1": [68729, 814703],
    "tuning/0": [754281],
    "held_out/0": [1500733],
}

SPLITS = {
    "train": [239684, 1195293, 68729, 814703],
    "tuning": [754281],
    "held_out": [1500733],
}

MEDS_TRAIN_0 = """
patient_id,time,code,numeric_value
239684,,EYE_COLOR//BROWN,
239684,,HEIGHT,175.271115221764
239684,"12/28/1980, 00:00:00",DOB,
239684,"05/11/2010, 17:41:51",ADMISSION//CARDIAC,
239684,"05/11/2010, 17:41:51",HR,102.6
239684,"05/11/2010, 17:41:51",TEMP,96.0
239684,"05/11/2010, 17:48:48",HR,105.1
239684,"05/11/2010, 17:48:48",TEMP,96.2
239684,"05/11/2010, 18:25:35",HR,113.4
239684,"05/11/2010, 18:25:35",TEMP,95.8
239684,"05/11/2010, 18:57:18",HR,112.6
239684,"05/11/2010, 18:57:18",TEMP,95.5
239684,"05/11/2010, 19:27:19",DISCHARGE,
1195293,,EYE_COLOR//BLUE,
1195293,,HEIGHT,164.6868838269085
1195293,"06/20/1978, 00:00:00",DOB,
1195293,"06/20/2010, 19:23:52",ADMISSION//CARDIAC,
1195293,"06/20/2010, 19:23:52",HR,109.0
1195293,"06/20/2010, 19:23:52",TEMP,100.0
1195293,"06/20/2010, 19:25:32",HR,114.1
1195293,"06/20/2010, 19:25:32",TEMP,100.0
1195293,"06/20/2010, 19:45:19",HR,119.8
1195293,"06/20/2010, 19:45:19",TEMP,99.9
1195293,"06/20/2010, 20:12:31",HR,112.5
1195293,"06/20/2010, 20:12:31",TEMP,99.8
1195293,"06/20/2010, 20:24:44",HR,107.7
1195293,"06/20/2010, 20:24:44",TEMP,100.0
1195293,"06/20/2010, 20:41:33",HR,107.5
1195293,"06/20/2010, 20:41:33",TEMP,100.4
1195293,"06/20/2010, 20:50:04",DISCHARGE,
"""

MEDS_TRAIN_1 = """
patient_id,time,code,numeric_value
68729,,EYE_COLOR//HAZEL,
68729,,HEIGHT,160.3953106166676
68729,"03/09/1978, 00:00:00",DOB,
68729,"05/26/2010, 02:30:56",ADMISSION//PULMONARY,
68729,"05/26/2010, 02:30:56",HR,86.0
68729,"05/26/2010, 02:30:56",TEMP,97.8
68729,"05/26/2010, 04:51:52",DISCHARGE,
814703,,EYE_COLOR//HAZEL,
814703,,HEIGHT,156.48559093209357
814703,"03/28/1976, 00:00:00",DOB,
814703,"02/05/2010, 05:55:39",ADMISSION//ORTHOPEDIC,
814703,"02/05/2010, 05:55:39",HR,170.2
814703,"02/05/2010, 05:55:39",TEMP,100.1
814703,"02/05/2010, 07:02:30",DISCHARGE,
"""

MEDS_TUNING_0 = """
patient_id,time,code,numeric_value
754281,,EYE_COLOR//BROWN,
754281,,HEIGHT,166.22261567137025
754281,"12/19/1988, 00:00:00",DOB,
754281,"01/03/2010, 06:27:59",ADMISSION//PULMONARY,
754281,"01/03/2010, 06:27:59",HR,142.0
754281,"01/03/2010, 06:27:59",TEMP,99.8
754281,"01/03/2010, 08:22:13",DISCHARGE,
"""

MEDS_HELD_OUT_0 = """
patient_id,time,code,numeric_value
1500733,,EYE_COLOR//BROWN,
1500733,,HEIGHT,158.60131573580904
1500733,"07/20/1986, 00:00:00",DOB,
1500733,"06/03/2010, 14:54:38",ADMISSION//ORTHOPEDIC,
1500733,"06/03/2010, 14:54:38",HR,91.4
1500733,"06/03/2010, 14:54:38",TEMP,100.0
1500733,"06/03/2010, 15:39:49",HR,84.4
1500733,"06/03/2010, 15:39:49",TEMP,100.3
1500733,"06/03/2010, 16:20:49",HR,90.1
1500733,"06/03/2010, 16:20:49",TEMP,100.1
1500733,"06/03/2010, 16:44:26",DISCHARGE,
"""

MEDS_SHARDS = parse_meds_csvs(
    {
        "train/0": MEDS_TRAIN_0,
        "train/1": MEDS_TRAIN_1,
        "tuning/0": MEDS_TUNING_0,
        "held_out/0": MEDS_HELD_OUT_0,
    }
)


MEDS_CODE_METADATA_CSV = """
code,code/n_occurrences,code/n_patients,values/n_occurrences,values/sum,values/sum_sqd,description,parent_codes
,44,4,28,3198.8389005974336,382968.28937288234,,
ADMISSION//CARDIAC,2,2,0,,,,
ADMISSION//ORTHOPEDIC,1,1,0,,,,
ADMISSION//PULMONARY,1,1,0,,,,
DISCHARGE,4,4,0,,,,
DOB,4,4,0,,,,
EYE_COLOR//BLUE,1,1,0,,,"Blue Eyes. Less common than brown.",
EYE_COLOR//BROWN,1,1,0,,,"Brown Eyes. The most common eye color.",
EYE_COLOR//HAZEL,2,2,0,,,"Hazel eyes. These are uncommon",
HEIGHT,4,4,4,656.8389005974336,108056.12937288235,,
HR,12,4,12,1360.5000000000002,158538.77,"Heart Rate",LOINC/8867-4
TEMP,12,4,12,1181.4999999999998,116373.38999999998,"Body Temperature",LOINC/8310-5
"""

MEDS_CODE_METADATA_SCHEMA = {
    "code": pl.Utf8,
    "code/n_occurrences": pl.UInt8,
    "code/n_patients": pl.UInt8,
    "values/n_occurrences": pl.UInt8,
    "values/n_patients": pl.UInt8,
    "values/sum": pl.Float32,
    "values/sum_sqd": pl.Float32,
    "values/n_ints": pl.UInt8,
    "values/min": pl.Float32,
    "values/max": pl.Float32,
    "description": pl.Utf8,
    "parent_codes": pl.Utf8,
    "code/vocab_index": pl.UInt8,
}


def parse_shards_yaml(yaml_str: str, **schema_updates) -> pl.DataFrame:
    schema = {**MEDS_PL_SCHEMA, **schema_updates}
    return parse_meds_csvs(load_yaml(yaml_str, Loader=Loader), schema=schema)


def parse_code_metadata_csv(csv_str: str) -> pl.DataFrame:
    cols = csv_str.strip().split("\n")[0].split(",")
    schema = {col: dt for col, dt in MEDS_CODE_METADATA_SCHEMA.items() if col in cols}
    df = pl.read_csv(StringIO(csv_str), schema=schema)
    if "parent_codes" in cols:
        df = df.with_columns(pl.col("parent_codes").cast(pl.List(pl.Utf8)))
    return df


MEDS_CODE_METADATA = parse_code_metadata_csv(MEDS_CODE_METADATA_CSV)


def check_NRT_output(
    output_fp: Path,
    want_nrt: JointNestedRaggedTensorDict,
):
    assert output_fp.is_file(), f"Expected {output_fp} to exist."

    got_nrt = JointNestedRaggedTensorDict.load(output_fp)

    # assert got_nrt.schema == want_nrt.schema, (
    #    f"Expected the schema of the NRT at {output_fp} to be equal to the target.\n"
    #    f"Wanted:\n{want_nrt.schema}\n"
    #    f"Got:\n{got_nrt.schema}"
    # )

    want_tensors = want_nrt.tensors
    got_tensors = got_nrt.tensors

    assert got_tensors.keys() == want_tensors.keys(), (
        f"Expected the keys of the NRT at {output_fp} to be equal to the target.\n"
        f"Wanted:\n{list(want_tensors.keys())}\n"
        f"Got:\n{list(got_tensors.keys())}"
    )

    for k in want_tensors.keys():
        want_v = want_tensors[k]
        got_v = got_tensors[k]

        assert type(want_v) is type(got_v), (
            f"Expected tensor {k} of the NRT at {output_fp} to be of the same type as the target.\n"
            f"Wanted:\n{type(want_v)}\n"
            f"Got:\n{type(got_v)}"
        )

        if isinstance(want_v, list):
            assert len(want_v) == len(got_v), (
                f"Expected list {k} of the NRT at {output_fp} to be of the same length as the target.\n"
                f"Wanted:\n{len(want_v)}\n"
                f"Got:\n{len(got_v)}"
            )
            for i, (want_i, got_i) in enumerate(zip(want_v, got_v)):
                assert np.array_equal(want_i, got_i, equal_nan=True), (
                    f"Expected tensor {k}[{i}] of the NRT at {output_fp} to be equal to the target.\n"
                    f"Wanted:\n{want_i}\n"
                    f"Got:\n{got_i}"
                )
        else:
            assert np.array_equal(want_v, got_v, equal_nan=True), (
                f"Expected tensor {k} of the NRT at {output_fp} to be equal to the target.\n"
                f"Wanted:\n{want_v}\n"
                f"Got:\n{got_v}"
            )


def check_df_output(
    output_fp: Path,
    want_df: pl.DataFrame,
    check_column_order: bool = False,
    check_row_order: bool = True,
    **kwargs,
):
    assert output_fp.is_file(), f"Expected {output_fp} to exist."

    got_df = pl.read_parquet(output_fp, glob=False)
    assert_df_equal(
        want_df,
        got_df,
        (f"Expected the dataframe at {output_fp} to be equal to the target.\n"),
        check_column_order=check_column_order,
        check_row_order=check_row_order,
        **kwargs,
    )


@contextmanager
def input_MEDS_dataset(
    input_code_metadata: pl.DataFrame | str | None = None,
    input_shards: dict[str, pl.DataFrame] | None = None,
    input_shards_map: dict[str, list[int]] | None = None,
    input_splits_map: dict[str, list[int]] | None = None,
):
    with tempfile.TemporaryDirectory() as d:
        MEDS_dir = Path(d) / "MEDS_cohort"
        cohort_dir = Path(d) / "output_cohort"

        MEDS_data_dir = MEDS_dir / "data"
        MEDS_metadata_dir = MEDS_dir / "metadata"

        # Create the directories
        MEDS_data_dir.mkdir(parents=True)
        MEDS_metadata_dir.mkdir(parents=True)
        cohort_dir.mkdir(parents=True)

        # Write the shards map
        if input_shards_map is None:
            input_shards_map = SHARDS

        shards_fp = MEDS_metadata_dir / ".shards.json"
        shards_fp.write_text(json.dumps(input_shards_map))

        # Write the splits parquet file
        if input_splits_map is None:
            input_splits_map = SPLITS
        input_splits_as_df = defaultdict(list)
        for split_name, patient_ids in input_splits_map.items():
            input_splits_as_df["patient_id"].extend(patient_ids)
            input_splits_as_df["split"].extend([split_name] * len(patient_ids))
        input_splits_df = pl.DataFrame(input_splits_as_df)
        input_splits_fp = MEDS_metadata_dir / "patient_splits.parquet"
        input_splits_df.write_parquet(input_splits_fp, use_pyarrow=True)

        if input_shards is None:
            input_shards = MEDS_SHARDS

        # Write the shards
        for shard_name, df in input_shards.items():
            fp = MEDS_data_dir / f"{shard_name}.parquet"
            fp.parent.mkdir(parents=True, exist_ok=True)
            df.write_parquet(fp, use_pyarrow=True)

        code_metadata_fp = MEDS_metadata_dir / "codes.parquet"
        if input_code_metadata is None:
            input_code_metadata = MEDS_CODE_METADATA
        elif isinstance(input_code_metadata, str):
            input_code_metadata = parse_code_metadata_csv(input_code_metadata)
        input_code_metadata.write_parquet(code_metadata_fp, use_pyarrow=True)

        yield MEDS_dir, cohort_dir


def check_outputs(
    cohort_dir: Path,
    want_data: dict[str, pl.DataFrame] | None = None,
    want_metadata: dict[str, pl.DataFrame] | pl.DataFrame | None = None,
    assert_no_other_outputs: bool = True,
    outputs_from_cohort_dir: bool = False,
):
    if want_metadata is not None:
        if isinstance(want_metadata, pl.DataFrame):
            want_metadata = {"codes.parquet": want_metadata}
        metadata_root = cohort_dir if outputs_from_cohort_dir else cohort_dir / "metadata"
        for shard_name, want in want_metadata.items():
            if Path(shard_name).suffix == "":
                shard_name = f"{shard_name}.parquet"
            check_df_output(metadata_root / shard_name, want)

    if want_data:
        data_root = cohort_dir if outputs_from_cohort_dir else cohort_dir / "data"
        all_file_suffixes = set()
        for shard_name, want in want_data.items():
            if Path(shard_name).suffix == "":
                shard_name = f"{shard_name}.parquet"

            file_suffix = Path(shard_name).suffix
            all_file_suffixes.add(file_suffix)

            output_fp = data_root / f"{shard_name}"
            if file_suffix == ".parquet":
                check_df_output(output_fp, want)
            elif file_suffix == ".nrt":
                check_NRT_output(output_fp, want)
            else:
                raise ValueError(f"Unknown file suffix: {file_suffix}")

        if assert_no_other_outputs:
            all_outputs = []
            for suffix in all_file_suffixes:
                all_outputs.extend(list((data_root).glob(f"**/*{suffix}")))
            assert len(want_data) == len(all_outputs), (
                f"Want {len(want_data)} outputs, but found {len(all_outputs)}.\n"
                f"Found outputs: {[fp.relative_to(data_root) for fp in all_outputs]}\n"
            )


def single_stage_transform_tester(
    transform_script: str | Path,
    stage_name: str,
    transform_stage_kwargs: dict[str, str] | None,
    do_pass_stage_name: bool = False,
    do_use_config_yaml: bool = False,
    want_data: dict[str, pl.DataFrame] | None = None,
    want_metadata: pl.DataFrame | None = None,
    assert_no_other_outputs: bool = True,
    **input_data_kwargs,
):
    with input_MEDS_dataset(**input_data_kwargs) as (MEDS_dir, cohort_dir):
        pipeline_config_kwargs = {
            "input_dir": str(MEDS_dir.resolve()),
            "cohort_dir": str(cohort_dir.resolve()),
            "stages": [stage_name],
            "hydra.verbose": True,
        }

        if transform_stage_kwargs:
            pipeline_config_kwargs["stage_configs"] = {stage_name: transform_stage_kwargs}

        run_command_kwargs = {
            "script": transform_script,
            "hydra_kwargs": pipeline_config_kwargs,
            "test_name": f"Single stage transform: {stage_name}",
        }
        if do_use_config_yaml:
            run_command_kwargs["do_use_config_yaml"] = True
            run_command_kwargs["config_name"] = "preprocess"
        if do_pass_stage_name:
            run_command_kwargs["stage"] = stage_name
            run_command_kwargs["do_pass_stage_name"] = True

        # Run the transform
        stderr, stdout = run_command(**run_command_kwargs)

        try:
            check_outputs(cohort_dir, want_data=want_data, want_metadata=want_metadata)
        except Exception as e:
            raise AssertionError(
                f"Single stage transform {stage_name} failed.\n"
                f"Script stdout:\n{stdout}\n"
                f"Script stderr:\n{stderr}"
            ) from e


def multi_stage_transform_tester(
    transform_scripts: list[str | Path],
    stage_names: list[str],
    stage_configs: dict[str, str] | str | None,
    do_pass_stage_name: bool | dict[str, bool] = True,
    want_data: dict[str, pl.DataFrame] | None = None,
    want_metadata: pl.DataFrame | None = None,
    outputs_from_cohort_dir: bool = True,
    **input_data_kwargs,
):
    with input_MEDS_dataset(**input_data_kwargs) as (MEDS_dir, cohort_dir):
        match stage_configs:
            case None:
                stage_configs = {}
            case str():
                stage_configs = load_yaml(stage_configs, Loader=Loader)
            case dict():
                pass
            case _:
                raise ValueError(f"Unknown stage_configs type: {type(stage_configs)}")

        match do_pass_stage_name:
            case True:
                do_pass_stage_name = {stage_name: True for stage_name in stage_names}
            case False:
                do_pass_stage_name = {stage_name: False for stage_name in stage_names}
            case dict():
                pass
            case _:
                raise ValueError(f"Unknown do_pass_stage_name type: {type(do_pass_stage_name)}")

        pipeline_config_kwargs = {
            "input_dir": str(MEDS_dir.resolve()),
            "cohort_dir": str(cohort_dir.resolve()),
            "stages": stage_names,
            "stage_configs": stage_configs,
            "hydra.verbose": True,
        }

        script_outputs = {}
        n_stages = len(stage_names)
        for i, (stage, script) in enumerate(zip(stage_names, transform_scripts)):
            script_outputs[stage] = run_command(
                script=script,
                hydra_kwargs=pipeline_config_kwargs,
                do_use_config_yaml=True,
                config_name="preprocess",
                test_name=f"Multi stage transform {i}/{n_stages}: {stage}",
                stage_name=stage,
                do_pass_stage_name=do_pass_stage_name[stage],
            )

        check_outputs(
            cohort_dir,
            want_data=want_data,
            want_metadata=want_metadata,
            outputs_from_cohort_dir=outputs_from_cohort_dir,
            assert_no_other_outputs=False,  # this currently doesn't work due to metadata / data confusions.
        )



"""Tests the full end-to-end extraction process.

Set the bash env variable `DO_USE_LOCAL_SCRIPTS=1` to use the local py files, rather than the installed
scripts.
"""

import os

import rootutils

root = rootutils.setup_root(__file__, dotenv=True, pythonpath=True, cwd=True)

code_root = root / "src" / "MEDS_transforms"
extraction_root = code_root / "extract"

if os.environ.get("DO_USE_LOCAL_SCRIPTS", "0") == "1":
    SHARD_EVENTS_SCRIPT = extraction_root / "shard_events.py"
    SPLIT_AND_SHARD_SCRIPT = extraction_root / "split_and_shard_patients.py"
    CONVERT_TO_SHARDED_EVENTS_SCRIPT = extraction_root / "convert_to_sharded_events.py"
    MERGE_TO_MEDS_COHORT_SCRIPT = extraction_root / "merge_to_MEDS_cohort.py"
    EXTRACT_CODE_METADATA_SCRIPT = extraction_root / "extract_code_metadata.py"
    FINALIZE_DATA_SCRIPT = extraction_root / "finalize_MEDS_data.py"
    FINALIZE_METADATA_SCRIPT = extraction_root / "finalize_MEDS_metadata.py"
else:
    SHARD_EVENTS_SCRIPT = "MEDS_extract-shard_events"
    SPLIT_AND_SHARD_SCRIPT = "MEDS_extract-split_and_shard_patients"
    CONVERT_TO_SHARDED_EVENTS_SCRIPT = "MEDS_extract-convert_to_sharded_events"
    MERGE_TO_MEDS_COHORT_SCRIPT = "MEDS_extract-merge_to_MEDS_cohort"
    EXTRACT_CODE_METADATA_SCRIPT = "MEDS_extract-extract_code_metadata"
    FINALIZE_DATA_SCRIPT = "MEDS_extract-finalize_MEDS_data"
    FINALIZE_METADATA_SCRIPT = "MEDS_extract-finalize_MEDS_metadata"

import json
import tempfile
from io import StringIO
from pathlib import Path

import polars as pl
from meds import __version__ as MEDS_VERSION

from .utils import assert_df_equal, run_command

# Test data (inputs)

SUBJECTS_CSV = """
MRN,dob,eye_color,height
1195293,06/20/1978,BLUE,164.6868838269085
239684,12/28/1980,BROWN,175.271115221764
1500733,07/20/1986,BROWN,158.60131573580904
814703,03/28/1976,HAZEL,156.48559093209357
754281,12/19/1988,BROWN,166.22261567137025
68729,03/09/1978,HAZEL,160.3953106166676
"""

ADMIT_VITALS_CSV = """
patient_id,admit_date,disch_date,department,vitals_date,HR,temp
239684,"05/11/2010, 17:41:51","05/11/2010, 19:27:19",CARDIAC,"05/11/2010, 18:57:18",112.6,95.5
754281,"01/03/2010, 06:27:59","01/03/2010, 08:22:13",PULMONARY,"01/03/2010, 06:27:59",142.0,99.8
814703,"02/05/2010, 05:55:39","02/05/2010, 07:02:30",ORTHOPEDIC,"02/05/2010, 05:55:39",170.2,100.1
239684,"05/11/2010, 17:41:51","05/11/2010, 19:27:19",CARDIAC,"05/11/2010, 18:25:35",113.4,95.8
68729,"05/26/2010, 02:30:56","05/26/2010, 04:51:52",PULMONARY,"05/26/2010, 02:30:56",86.0,97.8
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 20:12:31",112.5,99.8
1500733,"06/03/2010, 14:54:38","06/03/2010, 16:44:26",ORTHOPEDIC,"06/03/2010, 16:20:49",90.1,100.1
239684,"05/11/2010, 17:41:51","05/11/2010, 19:27:19",CARDIAC,"05/11/2010, 17:48:48",105.1,96.2
239684,"05/11/2010, 17:41:51","05/11/2010, 19:27:19",CARDIAC,"05/11/2010, 17:41:51",102.6,96.0
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 19:25:32",114.1,100.0
1500733,"06/03/2010, 14:54:38","06/03/2010, 16:44:26",ORTHOPEDIC,"06/03/2010, 14:54:38",91.4,100.0
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 20:41:33",107.5,100.4
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 20:24:44",107.7,100.0
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 19:45:19",119.8,99.9
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 19:23:52",109.0,100.0
1500733,"06/03/2010, 14:54:38","06/03/2010, 16:44:26",ORTHOPEDIC,"06/03/2010, 15:39:49",84.4,100.3
"""

INPUT_METADATA_FILE = """
lab_code,title,loinc
HR,Heart Rate,8867-4
temp,Body Temperature,8310-5
"""

DEMO_METADATA_FILE = """
eye_color,description
BROWN,"Brown Eyes. The most common eye color."
BLUE,"Blue Eyes. Less common than brown."
HAZEL,"Hazel eyes. These are uncommon"
GREEN,"Green eyes. These are rare."
"""

EVENT_CFGS_YAML = """
subjects:
  patient_id_col: MRN
  eye_color:
    code:
      - EYE_COLOR
      - col(eye_color)
    time: null
  height:
    code: HEIGHT
    time: null
    numeric_value: height
  dob:
    code: DOB
    time: col(dob)
    time_format: "%m/%d/%Y"
admit_vitals:
  admissions:
    code:
      - ADMISSION
      - col(department)
    time: col(admit_date)
    time_format: "%m/%d/%Y, %H:%M:%S"
  discharge:
    code: DISCHARGE
    time: col(disch_date)
    time_format: "%m/%d/%Y, %H:%M:%S"
  HR:
    code: HR
    time: col(vitals_date)
    time_format: "%m/%d/%Y, %H:%M:%S"
    numeric_value: HR
  temp:
    code: TEMP
    time: col(vitals_date)
    time_format: "%m/%d/%Y, %H:%M:%S"
    numeric_value: temp
"""

# Test data (expected outputs) -- ALL OF THIS MAY CHANGE IF THE SEED OR DATA CHANGES
EXPECTED_SPLITS = {
    "train/0": [239684, 1195293],
    "train/1": [68729, 814703],
    "tuning/0": [754281],
    "held_out/0": [1500733],
}

PATIENT_SPLITS_DF = pl.DataFrame(
    {
        "patient_id": [239684, 1195293, 68729, 814703, 754281, 1500733],
        "split": ["train", "train", "train", "train", "tuning", "held_out"],
    }
)


def get_expected_output(df: str) -> pl.DataFrame:
    return (
        pl.read_csv(source=StringIO(df))
        .select(
            "patient_id",
            pl.col("time").str.strptime(pl.Datetime, "%m/%d/%Y, %H:%M:%S").alias("time"),
            pl.col("code"),
            "numeric_value",
        )
        .sort(by=["patient_id", "time"])
    )


MEDS_OUTPUT_TRAIN_0_SUBJECTS = """
patient_id,time,code,numeric_value
239684,,EYE_COLOR//BROWN,
239684,,HEIGHT,175.271115221764
239684,"12/28/1980, 00:00:00",DOB,
1195293,,EYE_COLOR//BLUE,
1195293,,HEIGHT,164.6868838269085
1195293,"06/20/1978, 00:00:00",DOB,
"""

MEDS_OUTPUT_TRAIN_0_ADMIT_VITALS = """
patient_id,time,code,numeric_value
239684,"05/11/2010, 17:41:51",ADMISSION//CARDIAC,
239684,"05/11/2010, 17:41:51",HR,102.6
239684,"05/11/2010, 17:41:51",TEMP,96.0
239684,"05/11/2010, 17:48:48",HR,105.1
239684,"05/11/2010, 17:48:48",TEMP,96.2
239684,"05/11/2010, 18:25:35",HR,113.4
239684,"05/11/2010, 18:25:35",TEMP,95.8
239684,"05/11/2010, 18:57:18",HR,112.6
239684,"05/11/2010, 18:57:18",TEMP,95.5
239684,"05/11/2010, 19:27:19",DISCHARGE,
1195293,"06/20/2010, 19:23:52",ADMISSION//CARDIAC,
1195293,"06/20/2010, 19:23:52",HR,109.0
1195293,"06/20/2010, 19:23:52",TEMP,100.0
1195293,"06/20/2010, 19:25:32",HR,114.1
1195293,"06/20/2010, 19:25:32",TEMP,100.0
1195293,"06/20/2010, 19:45:19",HR,119.8
1195293,"06/20/2010, 19:45:19",TEMP,99.9
1195293,"06/20/2010, 20:12:31",HR,112.5
1195293,"06/20/2010, 20:12:31",TEMP,99.8
1195293,"06/20/2010, 20:24:44",HR,107.7
1195293,"06/20/2010, 20:24:44",TEMP,100.0
1195293,"06/20/2010, 20:41:33",HR,107.5
1195293,"06/20/2010, 20:41:33",TEMP,100.4
1195293,"06/20/2010, 20:50:04",DISCHARGE,
"""

MEDS_OUTPUT_TRAIN_1_SUBJECTS = """
patient_id,time,code,numeric_value
68729,,EYE_COLOR//HAZEL,
68729,,HEIGHT,160.3953106166676
68729,"03/09/1978, 00:00:00",DOB,
814703,,EYE_COLOR//HAZEL,
814703,,HEIGHT,156.48559093209357
814703,"03/28/1976, 00:00:00",DOB,
"""

MEDS_OUTPUT_TRAIN_1_ADMIT_VITALS = """
patient_id,time,code,numeric_value
68729,"05/26/2010, 02:30:56",ADMISSION//PULMONARY,
68729,"05/26/2010, 02:30:56",HR,86.0
68729,"05/26/2010, 02:30:56",TEMP,97.8
68729,"05/26/2010, 04:51:52",DISCHARGE,
814703,"02/05/2010, 05:55:39",ADMISSION//ORTHOPEDIC,
814703,"02/05/2010, 05:55:39",HR,170.2
814703,"02/05/2010, 05:55:39",TEMP,100.1
814703,"02/05/2010, 07:02:30",DISCHARGE,
"""

MEDS_OUTPUT_TUNING_0_SUBJECTS = """
patient_id,time,code,numeric_value
754281,,EYE_COLOR//BROWN,
754281,,HEIGHT,166.22261567137025
754281,"12/19/1988, 00:00:00",DOB,
"""

MEDS_OUTPUT_TUNING_0_ADMIT_VITALS = """
patient_id,time,code,numeric_value
754281,"01/03/2010, 06:27:59",ADMISSION//PULMONARY,
754281,"01/03/2010, 06:27:59",HR,142.0
754281,"01/03/2010, 06:27:59",TEMP,99.8
754281,"01/03/2010, 08:22:13",DISCHARGE,
"""

MEDS_OUTPUT_HELD_OUT_0_SUBJECTS = """
patient_id,time,code,numeric_value
1500733,,EYE_COLOR//BROWN,
1500733,,HEIGHT,158.60131573580904
1500733,"07/20/1986, 00:00:00",DOB,
"""

MEDS_OUTPUT_HELD_OUT_0_ADMIT_VITALS = """
patient_id,time,code,numeric_value
1500733,"06/03/2010, 14:54:38",ADMISSION//ORTHOPEDIC,
1500733,"06/03/2010, 14:54:38",HR,91.4
1500733,"06/03/2010, 14:54:38",TEMP,100.0
1500733,"06/03/2010, 15:39:49",HR,84.4
1500733,"06/03/2010, 15:39:49",TEMP,100.3
1500733,"06/03/2010, 16:20:49",HR,90.1
1500733,"06/03/2010, 16:20:49",TEMP,100.1
1500733,"06/03/2010, 16:44:26",DISCHARGE,
"""

MEDS_OUTPUT_CODE_METADATA_FILE = """
code,description,parent_codes
"""

MEDS_OUTPUT_DATASET_METADATA_JSON = {
    "dataset_name": "TEST",
    "dataset_version": "1.0",
    "etl_name": "MEDS_transforms",
    # "etl_version": None,  # We don't test this as it changes with the commits.
    "meds_version": MEDS_VERSION,
}

SUB_SHARDED_OUTPUTS = {
    "train/0": {
        "subjects": MEDS_OUTPUT_TRAIN_0_SUBJECTS,
        "admit_vitals": MEDS_OUTPUT_TRAIN_0_ADMIT_VITALS,
    },
    "train/1": {
        "subjects": MEDS_OUTPUT_TRAIN_1_SUBJECTS,
        "admit_vitals": MEDS_OUTPUT_TRAIN_1_ADMIT_VITALS,
    },
    "tuning/0": {
        "subjects": MEDS_OUTPUT_TUNING_0_SUBJECTS,
        "admit_vitals": MEDS_OUTPUT_TUNING_0_ADMIT_VITALS,
    },
    "held_out/0": {
        "subjects": MEDS_OUTPUT_HELD_OUT_0_SUBJECTS,
        "admit_vitals": MEDS_OUTPUT_HELD_OUT_0_ADMIT_VITALS,
    },
}


MEDS_OUTPUTS = {
    "train/0": [MEDS_OUTPUT_TRAIN_0_SUBJECTS, MEDS_OUTPUT_TRAIN_0_ADMIT_VITALS],
    "train/1": [MEDS_OUTPUT_TRAIN_1_SUBJECTS, MEDS_OUTPUT_TRAIN_1_ADMIT_VITALS],
    "tuning/0": [MEDS_OUTPUT_TUNING_0_SUBJECTS, MEDS_OUTPUT_TUNING_0_ADMIT_VITALS],
    "held_out/0": [MEDS_OUTPUT_HELD_OUT_0_SUBJECTS, MEDS_OUTPUT_HELD_OUT_0_ADMIT_VITALS],
}


def test_extraction():
    with tempfile.TemporaryDirectory() as d:
        raw_cohort_dir = Path(d) / "raw_cohort"
        MEDS_cohort_dir = Path(d) / "MEDS_cohort"

        # Create the directories
        raw_cohort_dir.mkdir()
        MEDS_cohort_dir.mkdir()

        subjects_csv = raw_cohort_dir / "subjects.csv"
        admit_vitals_csv = raw_cohort_dir / "admit_vitals.csv"
        event_cfgs_yaml = raw_cohort_dir / "event_cfgs.yaml"

        demo_metadata_csv = raw_cohort_dir / "demo_metadata.csv"
        input_metadata_csv = raw_cohort_dir / "input_metadata.csv"

        # Write the CSV files
        subjects_csv.write_text(SUBJECTS_CSV.strip())
        admit_vitals_csv.write_text(ADMIT_VITALS_CSV.strip())
        demo_metadata_csv.write_text(DEMO_METADATA_FILE.strip())
        input_metadata_csv.write_text(INPUT_METADATA_FILE.strip())

        # Mix things up -- have one CSV be also in parquet format.
        admit_vitals_parquet = raw_cohort_dir / "admit_vitals.parquet"
        df = pl.read_csv(admit_vitals_csv)

        df.write_parquet(admit_vitals_parquet, use_pyarrow=True)

        # Write the event config YAML
        event_cfgs_yaml.write_text(EVENT_CFGS_YAML)

        # Run the extraction script
        #   1. Sub-shard the data (this will be a null operation in this case, but it is worth doing just in
        #      case.
        #   2. Collect the patient splits.
        #   3. Extract the events and sub-shard by patient.
        #   4. Merge to the final output.

        extraction_config_kwargs = {
            "input_dir": str(raw_cohort_dir.resolve()),
            "cohort_dir": str(MEDS_cohort_dir.resolve()),
            "event_conversion_config_fp": str(event_cfgs_yaml.resolve()),
            "stage_configs.split_and_shard_patients.split_fracs.train": 4 / 6,
            "stage_configs.split_and_shard_patients.split_fracs.tuning": 1 / 6,
            "stage_configs.split_and_shard_patients.split_fracs.held_out": 1 / 6,
            "stage_configs.shard_events.row_chunksize": 10,
            "stage_configs.split_and_shard_patients.n_patients_per_shard": 2,
            "hydra.verbose": True,
            "etl_metadata.dataset_name": "TEST",
            "etl_metadata.dataset_version": "1.0",
        }

        all_stderrs = []
        all_stdouts = []

        # Stage 1: Sub-shard the data
        stderr, stdout = run_command(SHARD_EVENTS_SCRIPT, extraction_config_kwargs, "shard_events")

        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        subsharded_dir = MEDS_cohort_dir / "shard_events"

        try:
            out_files = list(subsharded_dir.glob("**/*.parquet"))
            assert len(out_files) == 3, f"Expected 3 output files, got {len(out_files)}."

            # Checking specific out files:
            #   1. subjects.parquet
            subjects_out = subsharded_dir / "subjects" / "[0-6).parquet"
            assert subjects_out.is_file(), f"Expected {subjects_out} to exist. Files include {out_files}."

            assert_df_equal(
                pl.read_csv(subjects_csv),
                pl.read_parquet(subjects_out, glob=False),
                "Subjects should be equal after sub-sharding",
                check_column_order=False,
                check_row_order=False,
            )
        except AssertionError as e:
            full_stderr = "\n".join(all_stderrs)
            print("Sub-sharding failed")
            print(f"stderr:\n{full_stderr}")
            raise e

        #   2. admit_vitals.parquet
        df_chunks = []
        for chunk in ["[0-10)", "[10-16)"]:
            admit_vitals_chunk_fp = subsharded_dir / "admit_vitals" / f"{chunk}.parquet"
            assert admit_vitals_chunk_fp.is_file(), f"Expected {admit_vitals_chunk_fp} to exist."

            df_chunks.append(pl.read_parquet(admit_vitals_chunk_fp, glob=False))

        assert_df_equal(
            pl.read_csv(admit_vitals_csv),
            pl.concat(df_chunks),
            "Admit vitals should be equal after sub-sharding",
            check_column_order=False,
            check_row_order=False,
        )

        # Stage 2: Collect the patient splits
        stderr, stdout = run_command(
            SPLIT_AND_SHARD_SCRIPT,
            extraction_config_kwargs,
            "split_and_shard_patients",
        )

        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        try:
            shards_fp = MEDS_cohort_dir / "metadata" / ".shards.json"
            assert shards_fp.is_file(), f"Expected splits @ {str(shards_fp.resolve())} to exist."

            splits = json.loads(shards_fp.read_text())
            expected_keys = ["train/0", "train/1", "tuning/0", "held_out/0"]

            expected_keys_str = ", ".join(f"'{k}'" for k in expected_keys)
            got_keys_str = ", ".join(f"'{k}'" for k in splits.keys())

            assert set(splits.keys()) == set(expected_keys), (
                f"Expected splits to have keys {expected_keys_str}.\n" f"Got keys: {got_keys_str}"
            )

            assert splits == EXPECTED_SPLITS, (
                f"Expected splits to be {EXPECTED_SPLITS}, got {splits}. NOTE THIS MAY CHANGE IF THE SEED OR "
                "DATA CHANGES -- FAILURE HERE MAY BE JUST DUE TO A NON-DETERMINISTIC SPLIT AND THE TEST "
                "NEEDING TO BE UPDATED."
            )
        except AssertionError as e:
            print("Failed to split patients")
            print(f"stderr:\n{stderr}")
            print(f"stdout:\n{stdout}")
            raise e

        # Stage 3: Extract the events and sub-shard by patient
        stderr, stdout = run_command(
            CONVERT_TO_SHARDED_EVENTS_SCRIPT,
            extraction_config_kwargs,
            "convert_events",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        patient_subsharded_folder = MEDS_cohort_dir / "convert_to_sharded_events"
        assert patient_subsharded_folder.is_dir(), f"Expected {patient_subsharded_folder} to be a directory."

        for split, expected_outputs in SUB_SHARDED_OUTPUTS.items():
            for prefix, expected_df_L in expected_outputs.items():
                if not isinstance(expected_df_L, list):
                    expected_df_L = [expected_df_L]

                expected_df = pl.concat([get_expected_output(df) for df in expected_df_L])

                fps = list((patient_subsharded_folder / split / prefix).glob("*.parquet"))
                assert len(fps) > 0

                # We add a "unique" here as there may be some duplicates across the row-group sub-shards.
                got_df = pl.concat([pl.read_parquet(fp, glob=False) for fp in fps]).unique()
                try:
                    assert_df_equal(
                        expected_df,
                        got_df,
                        f"Expected output for split {split}/{prefix} to be equal to the expected output.",
                        check_column_order=False,
                        check_row_order=False,
                    )
                except AssertionError as e:
                    print(f"Failed on split {split}/{prefix}")
                    print(f"stderr:\n{stderr}")
                    print(f"stdout:\n{stdout}")
                    raise e

        # Stage 4: Merge to the final output
        stderr, stdout = run_command(
            MERGE_TO_MEDS_COHORT_SCRIPT,
            extraction_config_kwargs,
            "merge_to_MEDS_cohort",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        full_stderr = "\n".join(all_stderrs)
        full_stdout = "\n".join(all_stdouts)

        # Check the final output
        output_folder = MEDS_cohort_dir / "merge_to_MEDS_cohort"
        try:
            for split, expected_df_L in MEDS_OUTPUTS.items():
                if not isinstance(expected_df_L, list):
                    expected_df_L = [expected_df_L]

                expected_df = pl.concat([get_expected_output(df) for df in expected_df_L])

                fp = output_folder / f"{split}.parquet"
                assert fp.is_file(), f"Expected {fp} to exist.\nstderr:\n{stderr}\nstdout:\n{stdout}"

                got_df = pl.read_parquet(fp, glob=False)
                assert_df_equal(
                    expected_df,
                    got_df,
                    f"Expected output for split {split} to be equal to the expected output.",
                    check_column_order=False,
                    check_row_order=False,
                )

                assert got_df["patient_id"].is_sorted(), f"Patient IDs should be sorted for split {split}."
                for subj in splits[split]:
                    got_df_subj = got_df.filter(pl.col("patient_id") == subj)
                    assert got_df_subj[
                        "time"
                    ].is_sorted(), f"Times should be sorted for patient {subj} in split {split}."

        except AssertionError as e:
            print(f"Failed on split {split}")
            print(f"stderr:\n{full_stderr}")
            print(f"stdout:\n{full_stdout}")
            raise e

        # Stage 6: Extract code metadata
        stderr, stdout = run_command(
            EXTRACT_CODE_METADATA_SCRIPT,
            extraction_config_kwargs,
            "extract_code_metadata",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        full_stderr = "\n".join(all_stderrs)
        full_stdout = "\n".join(all_stdouts)

        output_file = MEDS_cohort_dir / "extract_code_metadata" / "codes.parquet"
        assert (
            not output_file.is_file()
        ), f"Expected {output_file} to not  exist: stderr:\n{stderr}\nstdout:\n{stdout}"

        # Stage 7: Finalize the MEDS data
        stderr, stdout = run_command(
            FINALIZE_DATA_SCRIPT,
            extraction_config_kwargs,
            "finalize_MEDS_data",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        full_stderr = "\n".join(all_stderrs)
        full_stdout = "\n".join(all_stdouts)

        # Check the final output
        output_folder = MEDS_cohort_dir / "data"
        try:
            for split, expected_df_L in MEDS_OUTPUTS.items():
                if not isinstance(expected_df_L, list):
                    expected_df_L = [expected_df_L]

                expected_df = pl.concat([get_expected_output(df) for df in expected_df_L]).with_columns(
                    pl.col("numeric_value").cast(pl.Float32)
                )

                fp = output_folder / f"{split}.parquet"
                assert fp.is_file(), f"Expected {fp} to exist."

                got_df = pl.read_parquet(fp, glob=False)
                assert_df_equal(
                    expected_df,
                    got_df,
                    f"Expected output for split {split} to be equal to the expected output.",
                    check_column_order=False,
                    check_row_order=False,
                )

                assert got_df["patient_id"].is_sorted(), f"Patient IDs should be sorted for split {split}."
                for subj in splits[split]:
                    got_df_subj = got_df.filter(pl.col("patient_id") == subj)
                    assert got_df_subj[
                        "time"
                    ].is_sorted(), f"Times should be sorted for patient {subj} in split {split}."

        except AssertionError as e:
            print(f"Failed on split {split}")
            print(f"stderr:\n{full_stderr}")
            print(f"stdout:\n{full_stdout}")
            raise e

        # Stage 8: Finalize the metadata
        stderr, stdout = run_command(
            FINALIZE_METADATA_SCRIPT,
            extraction_config_kwargs,
            "finalize_metadata",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        full_stderr = "\n".join(all_stderrs)
        full_stdout = "\n".join(all_stdouts)

        # Check code metadata
        output_file = MEDS_cohort_dir / "metadata" / "codes.parquet"
        assert output_file.is_file(), f"Expected {output_file} to exist: stderr:\n{stderr}\nstdout:\n{stdout}"

        got_df = pl.read_parquet(output_file, glob=False, use_pyarrow=True)

        want_df = pl.read_csv(source=StringIO(MEDS_OUTPUT_CODE_METADATA_FILE)).with_columns(
            pl.col("code"),
            pl.col("parent_codes").cast(pl.List(pl.Utf8)),
        )

        # We collapse the list type as it throws an error in the assert_df_equal otherwise
        got_df = got_df.with_columns(pl.col("parent_codes").list.join("||"))
        want_df = want_df.with_columns(pl.col("parent_codes").list.join("||"))

        assert_df_equal(
            want=want_df,
            got=got_df,
            msg=f"Finalized code metadata differs:\nstderr:\n{stderr}\nstdout:\n{stdout}",
            check_column_order=False,
            check_row_order=False,
        )

        # Check dataset metadata
        output_file = MEDS_cohort_dir / "metadata" / "dataset.json"
        assert output_file.is_file(), f"Expected {output_file} to exist: stderr:\n{stderr}\nstdout:\n{stdout}"

        got_json = json.loads(output_file.read_text())
        assert "etl_version" in got_json, "Expected 'etl_version' to be in the dataset metadata."
        got_json.pop("etl_version")  # We don't test this as it changes with the commits.
        assert got_json == MEDS_OUTPUT_DATASET_METADATA_JSON, f"Dataset metadata differs: {got_json}"

        # Check the splits parquet
        output_file = MEDS_cohort_dir / "metadata" / "patient_splits.parquet"
        assert output_file.is_file(), f"Expected {output_file} to exist: stderr:\n{stderr}\nstdout:\n{stdout}"

        got_df = pl.read_parquet(output_file, glob=False, use_pyarrow=True)
        assert_df_equal(
            PATIENT_SPLITS_DF,
            got_df,
            "Patient splits should be equal to the expected splits.",
            check_column_order=False,
            check_row_order=False,
        )



#!/usr/bin/env python
"""A polars-to-polars transformation function for filtering patients by sequence length."""
from collections.abc import Callable
from functools import partial

import hydra
import polars as pl
from loguru import logger
from omegaconf import DictConfig

from MEDS_transforms import PREPROCESS_CONFIG_YAML
from MEDS_transforms.mapreduce.mapper import map_over


def filter_patients_by_num_measurements(df: pl.LazyFrame, min_measurements_per_patient: int) -> pl.LazyFrame:
    """Filters patients by the number of measurements they have.

    Args:
        df: The input DataFrame.
        min_measurements_per_patient: The minimum number of measurements a patient must have to be included.

    Returns:
        The filtered DataFrame.

    Examples:
        >>> df = pl.DataFrame({
        ...     "patient_id": [1, 1, 1, 2, 2, 3],
        ...     "time": [1, 2, 1, 1, 2, 1],
        ... })
        >>> filter_patients_by_num_measurements(df, 1)
        shape: (6, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 3          ┆ 1    │
        └────────────┴──────┘
        >>> filter_patients_by_num_measurements(df, 2)
        shape: (5, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        └────────────┴──────┘
        >>> filter_patients_by_num_measurements(df, 3)
        shape: (3, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        └────────────┴──────┘
        >>> filter_patients_by_num_measurements(df, 4)
        shape: (0, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        └────────────┴──────┘
        >>> filter_patients_by_num_measurements(df, 2.2)
        Traceback (most recent call last):
            ...
        TypeError: min_measurements_per_patient must be an integer; got <class 'float'> 2.2
    """
    if not isinstance(min_measurements_per_patient, int):
        raise TypeError(
            f"min_measurements_per_patient must be an integer; got {type(min_measurements_per_patient)} "
            f"{min_measurements_per_patient}"
        )

    return df.filter(pl.col("time").count().over("patient_id") >= min_measurements_per_patient)


def filter_patients_by_num_events(df: pl.LazyFrame, min_events_per_patient: int) -> pl.LazyFrame:
    """Filters patients by the number of events (unique timepoints) they have.

    Args:
        df: The input DataFrame.
        min_events_per_patient: The minimum number of events a patient must have to be included.

    Returns:
        The filtered DataFrame.

    Examples:
        >>> df = pl.DataFrame({
        ...     "patient_id": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4],
        ...     "time": [1, 1, 1, 1, 2, 1, 1, 2, 3, None, None, 1, 2, 3],
        ... })
        >>> filter_patients_by_num_events(df, 1)
        shape: (14, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 1    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 2          ┆ 1    │
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_patients_by_num_events(df, 2)
        shape: (11, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 2          ┆ 1    │
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_patients_by_num_events(df, 3)
        shape: (8, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_patients_by_num_events(df, 4)
        shape: (5, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_patients_by_num_events(df, 5)
        shape: (0, 2)
        ┌────────────┬──────┐
        │ patient_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        └────────────┴──────┘
        >>> filter_patients_by_num_events(df, 2.2)
        Traceback (most recent call last):
            ...
        TypeError: min_events_per_patient must be an integer; got <class 'float'> 2.2
    """
    if not isinstance(min_events_per_patient, int):
        raise TypeError(
            f"min_events_per_patient must be an integer; got {type(min_events_per_patient)} "
            f"{min_events_per_patient}"
        )

    return df.filter(pl.col("time").n_unique().over("patient_id") >= min_events_per_patient)


def filter_patients_fntr(stage_cfg: DictConfig) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    compute_fns = []
    if stage_cfg.min_measurements_per_patient:
        logger.info(
            f"Filtering patients with fewer than {stage_cfg.min_measurements_per_patient} measurements "
            "(observations of any kind)."
        )
        compute_fns.append(
            partial(
                filter_patients_by_num_measurements,
                min_measurements_per_patient=stage_cfg.min_measurements_per_patient,
            )
        )
    if stage_cfg.min_events_per_patient:
        logger.info(
            f"Filtering patients with fewer than {stage_cfg.min_events_per_patient} events "
            "(unique timepoints)."
        )
        compute_fns.append(
            partial(filter_patients_by_num_events, min_events_per_patient=stage_cfg.min_events_per_patient)
        )

    def fn(data: pl.LazyFrame) -> pl.LazyFrame:
        for compute_fn in compute_fns:
            data = compute_fn(data)
        return data

    return fn


@hydra.main(
    version_base=None, config_path=str(PREPROCESS_CONFIG_YAML.parent), config_name=PREPROCESS_CONFIG_YAML.stem
)
def main(cfg: DictConfig):
    """TODO."""

    map_over(cfg, compute_fn=filter_patients_fntr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import gzip
import warnings
from collections.abc import Callable
from enum import StrEnum
from pathlib import Path
from typing import TypeVar

import polars as pl
from loguru import logger


class SupportedFileFormats(StrEnum):
    """The supported file formats for dataframes we can read in, in priority order.

    The values of the enum are the allowed file suffix for the format.
    """

    PARQUET = ".parquet"
    CSV_GZ = ".csv.gz"
    CSV = ".csv"


def scan_csv_gz(fp: Path, **kwargs) -> pl.LazyFrame:
    with gzip.open(fp, mode="rb") as f:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            return pl.read_csv(f, **kwargs).lazy()


READERS = {
    SupportedFileFormats.PARQUET: pl.scan_parquet,
    SupportedFileFormats.CSV_GZ: scan_csv_gz,
    SupportedFileFormats.CSV: pl.scan_csv,
}


DF_T = TypeVar("DF_T")


def get_supported_fp(root_dir: Path, file_prefix: str | Path) -> tuple[Path, Callable[[Path], DF_T]]:
    """This function finds the best file path to read for a given root_dir and prefix.

    Args:
        root_dir: The root directory to search for files.
        file_prefix: The file prefix to search for.

    Raises:
        FileNotFoundError: If no files are found with the given prefix and an allowed suffix.

    Returns:
        The filepath with the matching prefix and the most preferred allowed suffix and an appropriate reader
        function for that file type.

    Examples:
        >>> from tempfile import TemporaryDirectory
        >>> df = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}, schema={"a": pl.UInt8, "b": pl.Int64})
        >>> with TemporaryDirectory() as tmpdir:
        ...     tmpdir = Path(tmpdir)
        ...     fp = tmpdir / "test.csv"
        ...     df.write_csv(fp)
        ...     fp, reader = get_supported_fp(tmpdir, "test")
        ...     print(str(fp.relative_to(tmpdir)), reader(fp).collect())
        test.csv shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 2   ┆ 5   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> with TemporaryDirectory() as tmpdir:
        ...     tmpdir = Path(tmpdir)
        ...     fp = tmpdir / "test.parquet"
        ...     csv_fp = tmpdir / "test.csv"
        ...     df.write_parquet(fp)
        ...     df.write_csv(csv_fp)
        ...     fp, reader = get_supported_fp(tmpdir, "test")
        ...     print(str(fp.relative_to(tmpdir)), reader(fp).collect())
        test.parquet shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ u8  ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 2   ┆ 5   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> import gzip
        >>> with TemporaryDirectory() as tmpdir:
        ...     tmpdir = Path(tmpdir)
        ...     fp = tmpdir / "test.csv.gz"
        ...     with gzip.open(fp, mode="wb") as f:
        ...         with warnings.catch_warnings():
        ...             warnings.simplefilter("ignore", category=UserWarning)
        ...             df.write_csv(f)
        ...     fp, reader = get_supported_fp(tmpdir, "test")
        ...     print(str(fp.relative_to(tmpdir)), reader(fp).collect())
        test.csv.gz shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 2   ┆ 5   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> with TemporaryDirectory() as tmpdir:
        ...     tmpdir = Path(tmpdir)
        ...     fp = tmpdir / "test.json"
        ...     df.write_json(fp)
        ...     get_supported_fp(tmpdir, "test") # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        FileNotFoundError: No files found with prefix: test and allowed suffixes
            ['.parquet', '.csv.gz', '.csv']...
    """

    for suffix in list(SupportedFileFormats):
        fp = root_dir / f"{file_prefix}{suffix.value}"
        if fp.exists():
            logger.debug(f"Found file: {str(fp.resolve())}")
            return fp, READERS[suffix]
    raise FileNotFoundError(
        f"No files found with prefix: {file_prefix} and allowed suffixes "
        f"{[x.value for x in SupportedFileFormats]} in root dir {str(root_dir.resolve())}"
    )



#!/usr/bin/env python

"""Performs pre-MEDS data wrangling for eICU.

See the docstring of `main` for more information.
"""

import gzip
from collections.abc import Callable
from datetime import datetime
from pathlib import Path

import hydra
import polars as pl
from loguru import logger
from omegaconf import DictConfig, OmegaConf

from MEDS_transforms.utils import get_shard_prefix, hydra_loguru_init, write_lazyframe

HEALTH_SYSTEM_STAY_ID = "patienthealthsystemstayid"
UNIT_STAY_ID = "patientunitstayid"
PATIENT_ID = "uniquepid"

# The end of year date, used for year-only timestamps in eICU. The time is set to midnight as we'll add a
# 24-hour time component from other columns in the data.
END_OF_YEAR = {"month": 12, "day": 31, "hour": 0, "minute": 0, "second": 0}


def load_raw_eicu_file(fp: Path, **kwargs) -> pl.LazyFrame:
    """Load a raw MIMIC file into a Polars DataFrame.

    Args:
        fp: The path to the MIMIC file.

    Returns:
        The Polars DataFrame containing the MIMIC data.
    """

    with gzip.open(fp, mode="rb") as f:
        return pl.read_csv(f, infer_schema_length=100000000, **kwargs).lazy()


def check_timestamps_agree(df: pl.LazyFrame, pseudotime_col: pl.Expr, given_24htime_col: str):
    """Checks that the time-of-day portions agree between the pseudotime and given columns.

    Raises a `ValueError` if the times don't match within a minute.

    Args:
        TODO
    """
    expected_time = pl.col(given_24htime_col).str.strptime(pl.Time, "%H:%M:%S")

    # The use of `.dt.combine` here re-sets the "time-of-day" of the pseudotime_col column
    time_deltas_min = (pseudotime_col - pseudotime_col.dt.combine(expected_time)).dt.total_minutes()

    # Check that the time deltas are all within 1 minute
    logger.info(
        "Checking that stated 24h times are consistent given offsets between {pseudotime_col.name} and "
        f"{given_24htime_col}..."
    )
    max_time_deltas_min = df.select(time_deltas_min.abs().max()).collect().item()
    if max_time_deltas_min > 1:
        raise ValueError(
            f"Max number of minutes between {pseudotime_col.name} and {given_24htime_col} is "
            f"{max_time_deltas_min}. Should be <= 1."
        )


def process_patient(df: pl.LazyFrame, hospital_df: pl.LazyFrame) -> pl.LazyFrame:
    """Takes the patient table and converts it to a form that includes timestamps.

    As eICU stores only offset times, note here that we add a CONSTANT TIME ACROSS ALL PATIENTS for the true
    timestamp of their health system admission. This is acceptable because in eICU ONLY RELATIVE TIME
    DIFFERENCES ARE MEANINGFUL, NOT ABSOLUTE TIMES.

    The output of this process is ultimately converted to events via the `patient` key in the
    `configs/event_configs.yaml` file.
    """

    hospital_discharge_pseudotime = pl.datetime(
        year=pl.col("hospitaldischargeyear"), **END_OF_YEAR
    ).dt.combine(pl.col("hospitaldischargetime24").str.strptime(pl.Time, "%H:%M:%S"))

    unit_admit_pseudotime = hospital_discharge_pseudotime - pl.duration(
        minutes=pl.col("hospitaldischargeoffset")
    )

    unit_discharge_pseudotime = unit_admit_pseudotime + pl.duration(minutes=pl.col("unitdischargeoffset"))

    hospital_admit_pseudotime = unit_admit_pseudotime + pl.duration(minutes=pl.col("hospitaladmitoffset"))

    age_in_years = (
        pl.when(pl.col("age") == "> 89").then(90).otherwise(pl.col("age").cast(pl.UInt16, strict=False))
    )
    age_in_days = age_in_years * 365.25
    # We assume that the patient was born at the midpoint of the year as we don't know the actual birthdate
    pseudo_date_of_birth = unit_admit_pseudotime - pl.duration(days=(age_in_days - 365.25 / 2))

    # Check the times
    start = datetime.now()
    logger.info(
        "Checking that the 24h times are consistent. If this is extremely slow, consider refactoring to have "
        "only one `.collect()` call."
    )
    check_timestamps_agree(df, hospital_discharge_pseudotime, "hospitaldischargetime24")
    check_timestamps_agree(df, hospital_admit_pseudotime, "hospitaladmittime24")
    check_timestamps_agree(df, unit_admit_pseudotime, "unitadmittime24")
    check_timestamps_agree(df, unit_discharge_pseudotime, "unitdischargetime24")
    logger.info(f"Validated 24h times in {datetime.now() - start}")

    logger.warning("NOT validating the `unitVisitNumber` column as that isn't implemented yet.")

    logger.warning(
        "NOT SURE ABOUT THE FOLLOWING. Check with the eICU team:\n"
        "  - `apacheAdmissionDx` is not selected from the patients table as we grab it from `admissiondx`. "
        "Is this right?\n"
        "  - `admissionHeight` and `admissionWeight` are interpreted as **unit** admission height/weight, "
        "not hospital admission height/weight. Is this right?\n"
        "  - `age` is interpreted as the age at the time of the unit stay, not the hospital stay. "
        "Is this right?\n"
        "  - `What is the actual mean age for those > 89? Here we assume 90.\n"
        "  - Note that all the column names appear to be all in lowercase for the csv versions, vs. the docs"
    )

    return df.join(hospital_df, left_on="hospitalid", right_on="hospitalid", how="left").select(
        # 1. Static variables
        PATIENT_ID,
        "gender",
        pseudo_date_of_birth.alias("dateofbirth"),
        "ethnicity",
        # 2. Health system stay parameters
        HEALTH_SYSTEM_STAY_ID,
        "hospitalid",
        pl.col("numbedscategory").alias("hospitalnumbedscategory"),
        pl.col("teachingstatus").alias("hospitalteachingstatus"),
        pl.col("region").alias("hospitalregion"),
        # 2.1 Admission parameters
        hospital_admit_pseudotime.alias("hospitaladmittimestamp"),
        "hospitaladmitsource",
        # 2.2 Discharge parameters
        hospital_discharge_pseudotime.alias("hospitaldischargetimestamp"),
        "hospitaldischargelocation",
        "hospitaldischargestatus",
        # 3. Unit stay parameters
        UNIT_STAY_ID,
        "wardid",
        # 3.1 Admission parameters
        unit_admit_pseudotime.alias("unitadmittimestamp"),
        "unitadmitsource",
        "unitstaytype",
        pl.col("admissionheight").alias("unitadmissionheight"),
        pl.col("admissionweight").alias("unitadmissionweight"),
        # 3.2 Discharge parameters
        unit_discharge_pseudotime.alias("unitdischargetimestamp"),
        "unitdischargelocation",
        "unitdischargestatus",
        pl.col("dischargeweight").alias("unitdischargeweight"),
    )


def join_and_get_pseudotime_fntr(
    table_name: str,
    offset_col: str | list[str],
    pseudotime_col: str | list[str],
    output_data_cols: list[str] | None = None,
    warning_items: list[str] | None = None,
) -> Callable[[pl.LazyFrame, pl.LazyFrame], pl.LazyFrame]:
    """Returns a function that joins a dataframe to the `patient` table and adds pseudotimes.

    Also raises specified warning strings via the logger for uncertain columns.

    TODO
    """

    if output_data_cols is None:
        output_data_cols = []

    if isinstance(offset_col, str):
        offset_col = [offset_col]
    if isinstance(pseudotime_col, str):
        pseudotime_col = [pseudotime_col]

    if len(offset_col) != len(pseudotime_col):
        raise ValueError(
            "There must be the same number of `offset_col`s and `pseudotime_col`s specified. Got "
            f"{len(offset_col)} and {len(pseudotime_col)}, respectively."
        )

    def fn(df: pl.LazyFrame, patient_df: pl.LazyFrame) -> pl.LazyFrame:
        f"""Takes the {table_name} table and converts it to a form that includes pseudo-timestamps.

        The output of this process is ultimately converted to events via the `{table_name}` key in the
        `configs/event_configs.yaml` file.
        """

        pseudotimes = [
            (pl.col("unitadmittimestamp") + pl.duration(minutes=pl.col(offset))).alias(pseudotime)
            for pseudotime, offset in zip(pseudotime_col, offset_col)
        ]

        if warning_items:
            warning_lines = [
                f"NOT SURE ABOUT THE FOLLOWING for {table_name} table. Check with the eICU team:",
                *(f"  - {item}" for item in warning_items),
            ]
            logger.warning("\n".join(warning_lines))

        return df.join(patient_df, on=UNIT_STAY_ID, how="inner").select(
            HEALTH_SYSTEM_STAY_ID,
            UNIT_STAY_ID,
            *pseudotimes,
            *output_data_cols,
        )

    return fn


NEEDED_PATIENT_COLS = [UNIT_STAY_ID, HEALTH_SYSTEM_STAY_ID, "unitadmittimestamp"]


@hydra.main(version_base=None, config_path="configs", config_name="pre_MEDS")
def main(cfg: DictConfig):
    """Performs pre-MEDS data wrangling for eICU.

    Inputs are the raw eICU files, read from the `raw_cohort_dir` config parameter. Output files are either
    symlinked (if they are not modified) or written in processed form to the `MEDS_input_dir` config
    parameter. Hydra is used to manage configuration parameters and logging.

    Note that eICU has only a tentative ability to identify true relative admission times for even the same
    patient, as health system stay IDs are only temporally ordered at the *year* level. As such, to properly
    parse this dataset in a longitudinal form, you must do one of the following:
      1. Not operate at the level of patients at all, but instead at the level of health system stays, as
         individual events within a health system stay can be well ordered.
      2. Restrict the analysis to only patients who do not have multiple health system stays within a single
         year (as health system stays across years can be well ordered, provided we assume to distinct stays
         within a single health system cannot overlap).

    In this pipeline, we choose to operate at the level of health system stays, as this is the most general
    approach. The only downside is that we lose the ability to track individual patients across health system
    stays, and thus can only explore questions of limited longitudinal scope.

    We ignore the following tables for the given reasons:
      1. `admissiondrug`: This table is noted in the
         [documentation](https://eicu-crd.mit.edu/eicutables/admissiondrug/) as being "Extremely infrequently
         used".
      2. `apacheApsVar`: This table is a sort of "meta-table" that contains variables used to compute the
         APACHE score; we won't use these raw variables from this table, but instead will use the raw data.
      3. `apachePatientResult`: This table has pre-computed APACHE score variables; we won't use these and
         will use the raw data directly.
      4. `apachePredVar`: This table contains variables used to compute the APACHE score; we won't use these
         in favor of the raw data directly.
      5. `carePlanCareProvider`: This table contains information about the provider for given care-plan
         entries; however, as we can't link this table to the particular care-plan entries, we don't use it
         here. It also is not clear (to the author of this script; the eICU team may know more) how reliable
         the time-offsets are for this table as they merely denote when a provider was entered into the care
         plan.
      6. `customLab`: The documentation for this table is very sparse, so we skip it.
      7. `intakeOutput`: There are a number of significant warnings about duplicates, cumulative values, and
         more in the documentation for this table, so for now we skip it.
      8. `microLab`: We don't use this because the culture taken time != culture result time, so seeing this
         data would give a model an advantage over any possible real-world implementation. Plus, the docs say
         it is not well populated.
      9. `note`: This table is largely duplicated with structured data due to the fact that primarily
         narrative notes were removed due to PHI constraints (see the docs).

    There are other notes for this pipeline:
      1. Many fields here are, I believe, **lists**, not simple categoricals, and should be split and
         processed accordingly. This is not yet done.

    Args (all as part of the config file):
        raw_cohort_dir: The directory containing the raw eICU files.
        output_dir: The directory to write the processed files to.
    """

    hydra_loguru_init()

    table_preprocessors_config_fp = Path("./eICU_Example/configs/table_preprocessors.yaml")
    logger.info(f"Loading table preprocessors from {str(table_preprocessors_config_fp.resolve())}...")
    preprocessors = OmegaConf.load(table_preprocessors_config_fp)
    functions = {}
    for table_name, preprocessor_cfg in preprocessors.items():
        logger.info(f"  Adding preprocessor for {table_name}:\n{OmegaConf.to_yaml(preprocessor_cfg)}")
        functions[table_name] = join_and_get_pseudotime_fntr(table_name=table_name, **preprocessor_cfg)

    raw_cohort_dir = Path(cfg.raw_cohort_dir)
    MEDS_input_dir = Path(cfg.output_dir)

    patient_out_fp = MEDS_input_dir / "patient.parquet"

    if patient_out_fp.is_file():
        logger.info(f"Reloading processed patient df from {str(patient_out_fp.resolve())}")
        patient_df = pl.read_parquet(patient_out_fp, columns=NEEDED_PATIENT_COLS, use_pyarrow=True).lazy()
    else:
        logger.info("Processing patient table first...")

        hospital_fp = raw_cohort_dir / "hospital.csv.gz"
        patient_fp = raw_cohort_dir / "patient.csv.gz"
        logger.info(f"Loading {str(hospital_fp.resolve())}...")
        hospital_df = load_raw_eicu_file(
            hospital_fp, columns=["hospitalid", "numbedscategory", "teachingstatus", "region"]
        )
        logger.info(f"Loading {str(patient_fp.resolve())}...")
        raw_patient_df = load_raw_eicu_file(patient_fp)

        logger.info("Processing patient table...")
        patient_df = process_patient(raw_patient_df, hospital_df)
        write_lazyframe(patient_df, MEDS_input_dir / "patient.parquet")

    all_fps = [
        fp for fp in raw_cohort_dir.glob("*.csv.gz") if fp.name not in {"hospital.csv.gz", "patient.csv.gz"}
    ]

    unused_tables = {
        "admissiondrug",
        "apacheApsVar",
        "apachePatientResult",
        "apachePredVar",
        "carePlanCareProvider",
        "customLab",
        "intakeOutput",
        "microLab",
        "note",
    }

    for in_fp in all_fps:
        pfx = get_shard_prefix(raw_cohort_dir, in_fp)
        if pfx in unused_tables:
            logger.warning(f"Skipping {pfx} as it is not supported in this pipeline.")
            continue
        elif pfx not in functions:
            logger.warning(f"No function needed for {pfx}. For eICU, THIS IS UNEXPECTED")
            continue

        out_fp = MEDS_input_dir / f"{pfx}.parquet"

        if out_fp.is_file():
            print(f"Done with {pfx}. Continuing")
            continue

        out_fp.parent.mkdir(parents=True, exist_ok=True)

        fn = functions[pfx]

        st = datetime.now()
        logger.info(f"Processing {pfx}...")
        df = load_raw_eicu_file(in_fp)
        logger.info(f"  * Loaded raw {in_fp} in {datetime.now() - st}")
        processed_df = fn(df, patient_df)
        write_lazyframe(processed_df, out_fp)
        logger.info(f"  * Processed and wrote to {str(out_fp.resolve())} in {datetime.now() - st}")

    logger.info(f"Done! All dataframes processed and written to {str(MEDS_input_dir.resolve())}")


if __name__ == "__main__":
    main()
